#include "externalVector.h"
#include "subscript.h"
#include "coerce.h"

static SEXP
ScalarVector(SEXPTYPE mode, const primitiveElt_t* value)
{
    SEXP ans;
    switch (mode) {
    case REALSXP:
        return ScalarReal(value->numeric);
    case INTSXP:
        return ScalarInteger(value->integer);
    case LGLSXP:
        return ScalarLogical(value->logical);
    case CPLXSXP:
        return ScalarComplex(value->cmplx);
    case STRSXP:
        return ScalarString((SEXP)value->character);
    default:
        PROTECT((SEXP)value->any);
        ans = allocVector(VECSXP, 1);
        SET_VECTOR_ELT(ans, 0, (SEXP)value->any);
        UNPROTECT(1);
        return ans;
    }
}

static SEXP
externalVector_setLength(SEXP vec, SEXP xmethods,
                         SEXP value)
{
    SEXP x = GET_SLOT(vec, install("storage"));
    SEXP storage;
    int len = asInteger(value);
    storageMethods_t* methods = R_ExternalPtrAddr(xmethods);
    int nx = methods->size(x);
    SEXP names;

    if (len == nx)
        return vec;
    if (NAMED(vec))
        vec = duplicate(vec);
    PROTECT(vec);
    x = PROTECT(methods->resize(x, len));
    SET_SLOT(vec, install("storage"), x);
    UNPROTECT(1);
    x = GET_SLOT(vec, install("storage"));
    if (len > nx) {
        setExtElt_t setElt = methods->setExtElt;
        primitiveElt_t NA = methods->NA;
        int i, warn;

        storage = GET_STORAGE_EXT_PTR(x);
        for (i = nx; i < len; i++) {
            setElt(storage, methods, i, &NA, &warn);
        }
    }
    SET_SLOT(x, install("length"), value);
    names = GET_SLOT(vec, install("Names"));
    if (names != R_NilValue) {
        if (!isExternalVectorWithStorage(names)) {
            SEXP newnames = allocVector(STRSXP, len);
            int i;
            PROTECT(newnames);
            if (len < nx)
                nx = len;
            names = coerceVector(names, STRSXP);
            PROTECT(names);
            for (i = 0; i < nx; i++)
                SET_STRING_ELT(newnames, i, STRING_ELT(names, i));
            SET_SLOT(vec, install("Names"), newnames);
            UNPROTECT(2);
        } else {
            methods = getStorageMethods(names);
            x = GET_SLOT(names, install("storage"));
            x = PROTECT(methods->resize(x, len));
            SET_SLOT(names, install("storage"), x);
            UNPROTECT(1);
            x = GET_SLOT(names, install("storage"));
            if (len > nx) {
                setExtElt_t setElt = methods->setExtElt;
                primitiveElt_t elt;
                int i, warn;

                elt.character = R_BlankString;
                storage = GET_STORAGE_EXT_PTR(x);
                for (i = nx; i < len; i++) {
                    setElt(storage, methods, i, &elt, &warn);
                }
            }
            SET_SLOT(x, install("length"), value);
        }
    }
    UNPROTECT(1);
    return vec;
}

typedef SEXP (*AttrGetter)(SEXP x, SEXP data);

static SEXP
characterSubset(SEXP x, SEXP sub, int nsub)
{
    SEXP ans;
    int i;

    if (x == R_NilValue)
        return x;
    ans = PROTECT(allocVector(STRSXP, nsub));
    if (isExternalVectorWithStorage(x)) {
        storageMethods_t* methods = getStorageMethods(x);
        SEXP storage = GET_STORAGE_EXT_PTR(GET_SLOT(x, install("storage")));
        int warn = 0;
        int n = asInteger(GET_SLOT(GET_SLOT(x, install("storage")),
                                   install("length")));

        methods->characterSubset(storage, methods, sub, n, ans, &warn);
        CoercionWarning(warn);
    } else {
        const SEXPREC* subtype = getIndexType(sub);
        if (subtype != noneIndex && subtype != NAIndex) {
            x = PROTECT(coerceVector(x, STRSXP));
            if (subtype == allIndex) {
                for (i = 0; i < nsub; i++)
                    SET_STRING_ELT(ans, i, STRING_ELT(x, i));
            } else if (subtype == sliceIndex) {
                double* content = REAL(GET_SLOT(sub, install("content")));
                int start = (int) content[0];
                int stride = (int) content[2];
                for (i = 0; i < nsub; i++, start += stride)
                    SET_STRING_ELT(ans, i, STRING_ELT(x, start));
            } else {
                int* indx = INTEGER(sub);
                for (i = 0; i < nsub; i++) {
                    SET_STRING_ELT(ans, i, STRING_ELT(x, indx[i]));
                }
            }
            UNPROTECT(1);
        }
    }
    UNPROTECT(1);
    return ans;
}

static SEXP
characterSubassign(SEXP x, SEXP sub, int nsub, SEXP value)
{
    int i;

    if (x == R_NilValue)
        return x;
    
    if (isExternalVectorWithStorage(x)) {
        extSubassign_t subassign;
        int warn = 0;
        SEXP xstorage = GET_SLOT(x, install("storage"));
        storageMethods_t* xmethods = getStorageMethods(xstorage);
        int n = asInteger(GET_SLOT(GET_SLOT(x, install("storage")),
                                   install("length")));
        subassign = xmethods->characterSubassign;
        subassign(xstorage, xmethods, sub,
                  n, value, &warn);
        CoercionWarning(warn);
        return x;
    } else {
        const SEXPREC* subtype = getIndexType(sub);
        if (subtype != noneIndex && subtype != NAIndex) {
            int nx;
            x = PROTECT(coerceVector(x, STRSXP));
            nx = LENGTH(x);
            if (subtype == allIndex) {
                for (i = 0; i < nx; i++)
                    SET_STRING_ELT(x, i, STRING_ELT(value, i));
            } else if (subtype == sliceIndex) {
                double* content = REAL(GET_SLOT(sub, install("content")));
                int start = (int) content[0];
                int stride = (int) content[2];
                for (i = 0; i < nsub; i++, start += stride)
                    SET_STRING_ELT(x, start, STRING_ELT(value, i));
            } else {
                int* indx = INTEGER(sub);
                for (i = 0; i < nsub; i++) {
                    SET_STRING_ELT(x, indx[i], STRING_ELT(value, i));
                }
            }
            UNPROTECT(1);
        }
    }
    UNPROTECT(1);
    return x;
}

static SEXP
VectorSubset(const SEXPREC* x, storageMethods_t* methods, SEXP subscript)
{
    SEXP externalStorage = GET_SLOT((SEXP)x, install("storage"));
    SEXP storage = GET_STORAGE_EXT_PTR(externalStorage);
    int nx = INTEGER(GET_SLOT(externalStorage, install("length")))[0];
    SEXP names = GET_SLOT((SEXP)x, install("Names"));
    int warn = 0;
    SEXP ans;

    ans = methods->vectorSubset(storage, methods, subscript, nx, R_NilValue,
                                &warn);
    PROTECT(ans);
    CoercionWarning(warn);
    SET_NAMES(ans, characterSubset(names, subscript, LENGTH(ans)));
    UNPROTECT(1);
    return ans;
}

static SEXP
MatrixSubset(SEXP x, storageMethods_t* xmethods, SEXP subs,
             Rboolean drop)
{
    SEXP dim = GET_SLOT(x, install("Dim"));
    int nr = INTEGER(dim)[0];
    int nc = INTEGER(dim)[1];
    SEXP xstorage = GET_SLOT(x, install("storage"));
    SEXP subr = VECTOR_ELT(subs, 0);
    SEXP subc = VECTOR_ELT(subs, 1);
    int nrsub, ncsub;
    int warn = 0;
    int* idim;
    SEXP ans;

    PROTECT(subr);
    PROTECT(subc);

    xstorage = GET_STORAGE_EXT_PTR(xstorage);
    ans = xmethods->matrixSubset(xstorage, xmethods, subr, subc, nr, nc,
                                 R_NilValue, &warn);
    PROTECT(ans);
    CoercionWarning(warn);
    warn = 0;
    idim = INTEGER(GET_DIM(ans));
    nrsub = idim[0];
    ncsub = idim[1];

    if (nrsub >= 0 && ncsub >= 0) {
        SEXP dimnames, dimnamesnames, newdimnames;
        SEXP attr = PROTECT(allocVector(INTSXP, 2));

        INTEGER(attr)[0] = nrsub;
        INTEGER(attr)[1] = ncsub;
        SET_DIM(ans, attr);
        UNPROTECT(1);

        dimnames = GET_SLOT(x, install("DimNames"));
        dimnamesnames = getAttrib(dimnames, R_NamesSymbol);
        if (!isNull(dimnames)) {
            newdimnames = PROTECT(allocVector(VECSXP, 2));
            if (TYPEOF(dimnames) == VECSXP) {
                SET_VECTOR_ELT(newdimnames, 0,
                               characterSubset(VECTOR_ELT(dimnames, 0),
                                               subr, nrsub));
                SET_VECTOR_ELT(newdimnames, 1,
                               characterSubset(VECTOR_ELT(dimnames, 1),
                                               subc, ncsub));
            } else {
                SET_VECTOR_ELT(newdimnames, 0,
                               characterSubset(CAR(dimnames),
                                               subr, nrsub));
                SET_VECTOR_ELT(newdimnames, 1,
                               characterSubset(CADR(dimnames),
                                               subc, ncsub));
            }
            setAttrib(newdimnames, R_NamesSymbol, dimnamesnames);
            setAttrib(ans, R_DimNamesSymbol, newdimnames);
            UNPROTECT(1);
        }
    }

    if (drop)
        DropDims(ans);
    UNPROTECT(3);
    return ans;
}

static SEXP
ArraySubset(SEXP x, storageMethods_t* methods,
            SEXP namesMethods, SEXP subs, Rboolean drop)
{
    error("Array subsetting not supported now");
    return R_NilValue; /* -Wall */
}

static SEXP
externalSubset(SEXP x, SEXP xmethods, SEXP namesMethods,
               SEXP subs, SEXP drop)
{
    int nsubs = length(subs);
    storageMethods_t* methods = R_ExternalPtrAddr(xmethods);
    if (nsubs == 1) {
        int nx = asInteger(GET_SLOT(GET_SLOT(x,
                                             install("storage")),
                                    install("length")));
        int i;
        const SEXPREC* subtype;
        subs = VECTOR_ELT(subs, 0);
        subtype = getIndexType(subs);
        if (subtype == noneIndex)
            nsubs = 0;
        else if (subtype == NAIndex) {
            SEXPTYPE mode = methods->mode;
            SEXP ans;
            nsubs = asInteger(GET_SLOT(subs, install("length")));
            ans = allocVector(mode, 0);

            switch (mode) {
            case LGLSXP:
            case INTSXP:
                for (i = 0; i < nsubs; i++)
                    INTEGER(ans)[i] = NA_INTEGER;
                break;
            case REALSXP:
                for (i = 0; i < nsubs; i++)
                    REAL(ans)[i] = NA_REAL;
                break;
            case CPLXSXP:
                for (i = 0; i < nsubs; i++) {
                    COMPLEX(ans)[i].r = NA_REAL;
                    COMPLEX(ans)[i].i = NA_REAL;
                }
                break;
            case STRSXP:
                for (i = 0; i < nsubs; i++)
                    SET_STRING_ELT(ans, i, NA_STRING);
                break;
            default:
                for (i = 0; i < nsubs; i++)
                    SET_VECTOR_ELT(ans, i, R_NilValue);
            }
            return ans;
        } else if (subtype == allIndex) {
            nsubs = nx;
        } else if (subtype == sliceIndex)
            nsubs = (int) REAL(GET_SLOT(subs, install("content")))[1];
        else nsubs = LENGTH(subs);

        if (nsubs == 0) {
            SEXPTYPE mode = methods->mode;
            return allocVector(mode, 0);
        }
        return VectorSubset(x, methods, subs);
    }
    if (nsubs != length(GET_SLOT(x, install("Dim"))))
        error("incorrect number of dimensions");
    if (nsubs == 2)
        return MatrixSubset(x, methods, subs,
                            asLogical(drop));
    return ArraySubset(x, methods, namesMethods, subs,
                       asLogical(drop));
}

static void
extVectorSubassign(SEXP x, storageMethods_t* xmethods,
                   SEXP subscript, SEXP y,
                   storageMethods_t* ymethods, int* warn)
{
    SEXP xstorage = GET_STORAGE_EXT_PTR(x);
    SEXP ystorage = GET_STORAGE_EXT_PTR(y);
    int ny = INTEGER(GET_SLOT(y, install("length")))[0];
    int* indx;
    primitiveElt_t value;
    getExtElt_t getElt = ymethods->getExtElt;
    setExtElt_t setElt;
    int i = 0;
    
    FIND_SET_ELT(setElt, xmethods, ymethods->mode,
                 "incompatible types in vector assignment");
    indx = INTEGER(PROTECT(asPositiveIndex(subscript, ny)));
    for (i = 0; i < ny; i++) {
        int j = indx[i];
        if (j != NA_INTEGER) {
            getElt(ystorage, ymethods, i, &value, warn);
            setElt(xstorage, xmethods, j, &value, warn);
        }
    }
    UNPROTECT(1);
}

static SEXP
VectorSubassign(SEXP x, SEXP sxmethods,
                SEXP subscript, SEXP y,
                SEXP symethods, SEXP newnames, SEXP newlength,
                int yIsExtVec)
{
    const SEXPREC* substype = getIndexType(subscript);
    SEXP ptr;
    SEXP xstorage;
    SEXP ystorage = R_NilValue; /* -Wall */
    storageMethods_t* xmethods = R_ExternalPtrAddr(sxmethods);
    storageMethods_t* ymethods = 0; /* -Wall */
    int nx;
    int ns, ny, warn=0;

    if (yIsExtVec) {
        ystorage = GET_SLOT(y, install("storage"));
        ymethods = R_ExternalPtrAddr(symethods);
        ny = ymethods->size(ystorage);
    } else {
        ny = LENGTH(y);
    }

    if (NAMED(x))
        x = duplicate(x);
    PROTECT(x);
    externalVector_setLength(x, sxmethods, newlength);
    xstorage = GET_SLOT(x, install("storage"));
    nx = xmethods->size(xstorage);

    if (substype == noneIndex)
        ns = 0;
    else if (substype == NAIndex)
        ns = asInteger(GET_SLOT(subscript, install("length")));
    else if (substype == allIndex)
        ns = nx;
    else if (substype == sliceIndex)
        ns = (int) REAL(GET_SLOT(subscript, install("content")))[1];
    else ns = LENGTH(subscript);

    if (ns > 0 && ny == 0)
	    error("nothing to replace with");
	if (ns > 0 && ns % ny)
	    warning("number of items to replace is not a multiple of replacement length");

    ptr = GET_STORAGE_EXT_PTR(xstorage);

    if (yIsExtVec) {
        extVectorSubassign(xstorage, xmethods, subscript,
                           ystorage,
                           ymethods, &warn);
    } else {
        switch (TYPEOF(y)) {
        case LGLSXP:
            xmethods->logicalSubassign(ptr, xmethods,
                                       subscript, nx, y, &warn);
            break;
        case INTSXP:
            xmethods->integerSubassign(ptr, xmethods,
                                      subscript, nx, y, &warn);
            break;
        case REALSXP:
            xmethods->numericSubassign(ptr, xmethods,
                                       subscript, nx, y, &warn);
            break;
        case CPLXSXP:
            xmethods->complexSubassign(ptr, xmethods,
                                       subscript, nx, y, &warn);
            break;
        case STRSXP:
            xmethods->characterSubassign(ptr, xmethods,
                                         subscript, nx, y, &warn);
            break;
        default:
            error("can not handle arbitrary types");
            break;
        }
        CoercionWarning(warn);
        warn = 0;
    }
    
    if (newnames != R_NilValue) {
        SEXP names = GET_SLOT(x, install("Names"));

        if (names == R_NilValue) {
            names = PROTECT(allocVector(STRSXP, nx));
            SET_SLOT(x, install("Names"), names);
            names = GET_SLOT(x, install("Names"));
            UNPROTECT(1);
        }
        characterSubassign(names, subscript, ns, newnames);
    }
    UNPROTECT(1);
    return x;
}

static SEXP
MatrixSubassign(SEXP x, SEXP sxmethods, SEXP s,
                SEXP y, SEXP symethods,
                int yIsExtVec)
{
    int nr, nc, nrs, ncs, ny, n;
    storageMethods_t* xmethods;
    SEXP xstorage;
    SEXP ystorage;
    SEXP sr, sc;

    SEXP dim = GET_SLOT(x, install("Dim"));
    if (TYPEOF(dim) != INTSXP || LENGTH(dim) != 2)
        error("incorrect number of subscripts on external matrix");
    if (NAMED(x))
        x = duplicate(x);
    PROTECT(x);
    xstorage = GET_SLOT(x, install("storage"));
    nr = INTEGER(dim)[0];
    nc = INTEGER(dim)[1];
    if (yIsExtVec) {
        ystorage = GET_SLOT(y, install("storage"));
        ny = asInteger(GET_SLOT(ystorage, install("length")));
    } else {
        ystorage = R_NilValue; /* -Wall */
        ny = LENGTH(y);
    }

    sr = VECTOR_ELT(s, 0);
    sc = VECTOR_ELT(s, 1);

    nrs = getIndexLength(sr, nr);
    ncs = getIndexLength(sc, nc);
    n = nrs*ncs;

    if (n > 0 && ny == 0)
        error("nothing to replace with");
    if (n > 0 && n % ny)
        error("number of items to replace is not a multiple of replacement length");
    xstorage = GET_STORAGE_EXT_PTR(xstorage);
    xmethods = (storageMethods_t*) R_ExternalPtrAddr(sxmethods);
    if (yIsExtVec) {
        storageMethods_t* ymethods =
            (storageMethods_t*) R_ExternalPtrAddr(symethods);
        int* indxc;
        int* indxr;
        getExtElt_t getElt = ymethods->getExtElt;
        setExtMatrixElt_t setMatElt;
        primitiveElt_t elt;
        int xwarn = 0;
        int ywarn = 0;
        int j, k;

        FIND_SETMAT_ELT(setMatElt, xmethods, ymethods->mode,
                        "incompatible types in external matrix subset assignment");
        ystorage = GET_STORAGE_EXT_PTR(ystorage);
        sr = PROTECT(asPositiveIndex(sr, nrs));
        indxr = INTEGER(sr);
        sc = PROTECT(asPositiveIndex(sc, ncs));
        indxc = INTEGER(sc);
        for (j = 0, k = 1; j < ncs; j++) {
            int jj = indxc[j];

            if (jj != NA_INTEGER) {
                int i;
                for (i = 0; i < nrs; i++) {
                    int ii = indxr[i];
                    if (ii != NA_INTEGER) {
                        setMatElt(xstorage, xmethods, ii, jj,
                                  nr, nc, getElt(ystorage, ymethods,
                                                 k, &elt, &ywarn),
                                  &xwarn);
                        if (k < ny)
                            k++;
                        else k = 1;
                    }
                }
            }
        }
        CoercionWarning(ywarn);
        CoercionWarning(xwarn);
    } else {
        int warn = 0;
        switch (TYPEOF(y)) {
        case LGLSXP:
            xmethods->logicalMatrixSubassign(xstorage, xmethods, sr, sc,
                                             nr, nc, y, &warn);
            break;
        case INTSXP:
            xmethods->integerMatrixSubassign(xstorage, xmethods, sr, sc,
                                             nr, nc, y, &warn);
            break;
        case REALSXP:
            xmethods->numericMatrixSubassign(xstorage, xmethods, sr, sc,
                                             nr, nc, y, &warn);
            break;
        case CPLXSXP:
            xmethods->complexMatrixSubassign(xstorage, xmethods, sr, sc,
                                             nr, nc, y, &warn);
            break;
        case STRSXP:
            xmethods->characterMatrixSubassign(xstorage, xmethods, sr, sc,
                                               nr, nc, y, &warn);
            break;
        default:
            error("incompatible types in external matrix subset assignment");
        }
        CoercionWarning(warn);
    }
    UNPROTECT(1);
    return x;
}

static SEXP
ArraySubassign(SEXP x, SEXP xmethods, SEXP subs, SEXP y,
               SEXP ymethods, int yIsExtVec)
{
    error("Array subassigning not supported now");
    return x;
}

static SEXP
externalSubassign(SEXP x, SEXP xmethods,
                  SEXP subs, SEXP newnames,
                  SEXP newlength, SEXP y,
                  SEXP ymethods)
{
    int nsubs = LENGTH(subs);
    int yIsExtVec = TYPEOF(ymethods) == EXTPTRSXP;
    /*
     * If the external pointer for x and y are same, we should
     * make a copy of it and then do the assignment. For now we just
     * print an error message.
     */
    if (yIsExtVec &&
        GET_STORAGE_EXT_PTR(GET_SLOT(x, install("storage"))) ==
        GET_STORAGE_EXT_PTR(GET_SLOT(y, install("storage"))))
        error("at present can not subassign an external object to itself");
    switch (nsubs) {
    case 1:
        return VectorSubassign(x, xmethods,
                               VECTOR_ELT(subs, 0), y, ymethods,
                               newnames, newlength,
                               yIsExtVec);
    case 2:
        return MatrixSubassign(x, xmethods,
                               subs, y, ymethods,
                               yIsExtVec);
    default:
        return ArraySubassign(x, xmethods,
                              subs, y, ymethods,
                              yIsExtVec);
    }
}

/*
 *  The externalVector objects x and y have the same length and
 *  different types.
 *  Each element of y is set to the corresponding element of x after
 *  coercing it to the type of y.
 */
static SEXP
externalVector_asEach(SEXP x, SEXP xmethods,
                      SEXP y, SEXP ymethods)
{
    SEXP xstorage, ystorage;
    storageMethods_t* xmethodsptr;
    storageMethods_t* ymethodsptr;
    primitiveElt_t value;
    getExtElt_t getElt;
    setExtElt_t setElt;
    int i, n;
    int xwarn = 0;
    int ywarn = 0;

    x = GET_SLOT((SEXP)x, install("storage"));
    xstorage = GET_STORAGE_EXT_PTR(x);
    ystorage = GET_STORAGE_EXT_PTR(GET_SLOT(y, install("storage")));
    xmethodsptr = R_ExternalPtrAddr(xmethods);
    ymethodsptr = R_ExternalPtrAddr(ymethods);
    getElt = xmethodsptr->getExtElt;
    FIND_SET_ELT(setElt, ymethodsptr, xmethodsptr->mode,
                 NULL);
    n = asInteger(GET_SLOT((SEXP)x, install("length")));

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethodsptr, i, &value, &xwarn);
        setElt(ystorage, ymethodsptr, i, &value, &ywarn);
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    return y;
}

static SEXP
externalVector_lapply(SEXP X, SEXP xlen, SEXP R_fcall, SEXP rho)
{
    int n = asInteger(xlen);
    SEXP ans = R_NilValue;
    SEXPTYPE mode;
    storageMethods_t* methods;
    SEXP storage;
    getExtElt_t getElt;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    if (!isExternalVectorWithStorage(X) ||
        !(isLanguage(R_fcall) && length(R_fcall) > 1) ||
        !isEnvironment(rho))
        error("invalid argument");

    if (n < 1)
        return allocVector(VECSXP, 0);
    storage = GET_SLOT(X, install("storage"));
    methods = getStorageMethods(storage);
    storage = GET_STORAGE_EXT_PTR(storage);
    getElt = methods->getExtElt;
    mode = methods->mode;
    PROTECT(R_fcall = duplicate(R_fcall));
    PROTECT(ans = allocVector(VECSXP, n));
    if (mode == VECSXP) {
        for (i = 0; i < n; i++) {
            SEXP tmp;
            getElt(storage, methods, i, &elt, &warn);
            SETCADR(R_fcall, elt.any);
            tmp = eval(R_fcall, rho);
            SET_VECTOR_ELT(ans, i, tmp);
        }
    } else {
        SEXP scalar = allocVector(mode, 1);
        setRElt_t setElt = methods->setRElt;

        SETCADR(R_fcall, scalar);
        for (i = 0; i < n; i++) {
            SEXP tmp;
            getElt(storage, methods, i, &elt, &warn);
            setElt(scalar, 0, &elt);
            tmp = eval(R_fcall, rho);
            SET_VECTOR_ELT(ans, i, tmp);
        }
    }
    UNPROTECT(2);
    CoercionWarning(warn);
    return ans;
}

#if 0
static Rboolean
isExternalVector(SEXP obj, R_CConvertInfo *info,
                 R_toCConverter *el)
{
    SEXP call = PROTECT(lang3(install("is"),
                              x, mkString("externalVector")));
    SEXP ans = eval(call, R_GlobalEnv);
    UNPROTECT(1);
    return asLogical(ans);
}

static void*
convertStorageToR(SEXP obj, R_CConvertInfo *info,
                  R_toCConverter *el)
{
    storageMethods_t* methods = getStorageMethods(obj);
    SEXP single = getAttrib(obj, install("CSingle"));
    void* ans;
    if (single != R_NilValue && asLogical(single)) { /* float */
        SEXP storage = GET_SLOT(obj, install("storage"));
        int n = GET_SLOT(storage, install("length"));
        int i;
        for (i = 0; i < n; i++) {
            
        }
    } else { /* everything else */
        SEXPTYPE mode = methods->mode;
        SEXP Rvec;
        switch (mode) {
        case LGLSXP:
        case INTSXP:
        case REALSXP:
        case CPLXSXP:
            Rvec = VectorSubset(obj, methods, allIndex);
            switch (TYPEOF(Rvec)) {
            default:
                error("Problem with conversion: ");
            case LGLSXP:
                ans = LOGICAL(Rvec);
                break;
            case INTSXP:
                ans = INTEGER(Rvec);
                break;
            case REALSXP:
                ans = REAL(Rvec);
                break;
            case CPLXSXP:
                ans = COMPLEX(Rvec);
                break;
            }
            break;
        case STRSXP:
            if (info->Fort) { /* Fortran */
            } else {
                ans = R_alloc()
                    for (i = 0; i < n; i++) {
                    }
            }
        case VECSXP:
            error("can not pass ");
        default:
            error("Unknown mode");
        }
    }
    return ans;
}

static SEXP
convertRToStorage(void* value, SEXP arg, R_FromCConvertInfo *info,
                  R_toCConverter *el)
{
    
}
#endif
#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(externalSubset, 5),
    CALL_DEF(externalSubassign, 7),
    CALL_DEF(externalVector_asEach, 4),
    CALL_DEF(externalVector_setLength, 3),
    CALL_DEF(externalVector_lapply, 4),
   {NULL, NULL, 0},
};

static const R_ExternalMethodDef R_ExternalDef  [] = {
    {NULL, NULL, 0},
};

R_CallMethodDef*
local_Call_methods(int* nptr, const R_CallMethodDef* callDef)
{
    int n;
    R_CallMethodDef* ans;

    for (n = 0; callDef[n].name != NULL; n++) {;}
    *nptr = n;
    ans = Calloc(++n, R_CallMethodDef);
    memcpy(ans, callDef, n*sizeof(R_CallMethodDef));
    return ans;
}

R_ExternalMethodDef*
local_External_methods(int* nptr, const R_ExternalMethodDef* extDef)
{
    int n;
    R_ExternalMethodDef* ans;

    for (n = 0; extDef[n].name != NULL; n++) {;}
    *nptr = n;
    ans = Calloc(++n, R_ExternalMethodDef);
    memcpy(ans, extDef, n*sizeof(R_ExternalMethodDef));
    return ans;
}


const R_CallMethodDef*
externalVector_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}

typedef const R_CallMethodDef* (*Call_methods_t)(int*);
typedef const R_ExternalMethodDef* (*External_methods_t)(int*);

#define REGISTRATION_METHOD_NAME(type, module) \
    (module ## _ ## type ## _methods)
#define DEFINE_REGISTRATION_METHOD(type, module) \
    const R_ ## type ## MethodDef* REGISTRATION_METHOD_NAME(type, module)(int*)

DEFINE_REGISTRATION_METHOD(Call, allocators);
DEFINE_REGISTRATION_METHOD(Call, array);
DEFINE_REGISTRATION_METHOD(Call, externalVector);
DEFINE_REGISTRATION_METHOD(External, groups);
DEFINE_REGISTRATION_METHOD(Call, groups);
DEFINE_REGISTRATION_METHOD(Call, math);
DEFINE_REGISTRATION_METHOD(External, slice);
DEFINE_REGISTRATION_METHOD(Call, slice);
DEFINE_REGISTRATION_METHOD(Call, storage);
DEFINE_REGISTRATION_METHOD(Call, subscript);

static const Call_methods_t call_methods_table[] =
{
    &REGISTRATION_METHOD_NAME(Call, allocators),
    &REGISTRATION_METHOD_NAME(Call, array),
    &REGISTRATION_METHOD_NAME(Call, externalVector),
    &REGISTRATION_METHOD_NAME(Call, groups),
    &REGISTRATION_METHOD_NAME(Call, math),
    &REGISTRATION_METHOD_NAME(Call, slice),
    &REGISTRATION_METHOD_NAME(Call, storage),
    &REGISTRATION_METHOD_NAME(Call, subscript)
};

#define N_CALL_METHODS (sizeof(call_methods_table)/sizeof(Call_methods_t))

static const External_methods_t external_methods_table[] =
{
    &REGISTRATION_METHOD_NAME(External, groups),
    &REGISTRATION_METHOD_NAME(External, slice)
};

#define N_EXTERNAL_METHODS \
    (sizeof(external_methods_table)/sizeof(External_methods_t))

R_CMethodDef* storage_C_methods(int*);

void
R_init_externalVector(DllInfo *dll)
{
    int n[N_CALL_METHODS];
    int n_ext[N_EXTERNAL_METHODS];
    int i, nsum;
    const R_CallMethodDef* other_CallDef[N_CALL_METHODS];
    const R_ExternalMethodDef* other_ExternalDef[N_EXTERNAL_METHODS];
    R_ExternalMethodDef* all_ExternalDef;
    R_CallMethodDef* all_CallDef;
    R_CMethodDef* all_CDef = storage_C_methods(&i);

    for (i = 0, nsum = 0; i < N_CALL_METHODS; i++) {
        other_CallDef[i] = (call_methods_table[i])(&n[i]);
        nsum += n[i];
    }
    all_CallDef = Calloc(nsum+1, R_CallMethodDef);
    for (i = N_CALL_METHODS-1, nsum = 0; i > 0; i--) {
        memcpy(all_CallDef+nsum, other_CallDef[i],
               n[i]*sizeof(R_CallMethodDef));
        nsum += n[i];
    }
    memcpy(all_CallDef+nsum, other_CallDef[0],
           (n[0]+1)*sizeof(R_CallMethodDef));

    for (i = 0, nsum = 0; i < N_EXTERNAL_METHODS; i++) {
        other_ExternalDef[i] = (external_methods_table[i])(&n_ext[i]);
        nsum += n_ext[i];
    }
    all_ExternalDef = Calloc(nsum+1, R_ExternalMethodDef);
    for (i = N_EXTERNAL_METHODS-1, nsum = 0; i > 0; i--) {
        memcpy(all_ExternalDef+nsum, other_ExternalDef[i],
               n_ext[i]*sizeof(R_ExternalMethodDef));
        nsum += n_ext[i];
    }
    memcpy(all_ExternalDef+nsum, other_ExternalDef[0],
           (n_ext[0]+1)*sizeof(R_ExternalMethodDef));

    R_registerRoutines(dll, all_CDef, all_CallDef, NULL,
                       all_ExternalDef);
    Free(all_CDef);
    Free(all_CallDef);
    Free(all_ExternalDef);

#if 0
    R_addToCConverter(isExternalVector, converToR, NULL,
                      "externalVector",
                      "Convert externalVector data to R data");
    R_addToCConverter(isExternalVectorWithStorage, converStorageToR,
                      convertRToStorage,
                      "externalVectorWithStorage",
                      "Convert externalVector data to R data");
#endif
}
