#include <Rmath.h>
#include "storageMethods.h"
#include "coerce.h"

#define R_log log
#define R_log10 log10

#define logbase(x, base) (R_log(x)/R_log(base))

static void complex_div(Rcomplex *a, Rcomplex *b)
{
    double ratio, den;
    double abr, abi;

    if( (abr = b->r) < 0)
	abr = - abr;
    if( (abi = b->i) < 0)
	abi = - abi;
    if( abr <= abi ) {
#ifndef IEEE_754
	if(abi == 0) {
	    b->r = NA_REAL;
	    b->i = NA_REAL;
	    return;
	}
#endif
	ratio = b->r / b->i ;
	den = b->i * (1 + ratio*ratio);
	b->r = (a->r*ratio + a->i) / den;
	b->i = (a->i*ratio - a->r) / den;
    }
    else {
	ratio = b->i / b->r ;
	den = b->r * (1 + ratio*ratio);
	b->r = (a->r + a->i*ratio) / den;
	b->i = (a->i - a->r*ratio) / den;
    }
}

static void
z_log(Rcomplex *z)
{
    double tmp = z->i;
    z->i = atan2(tmp, z->r);
    z->r = R_log(hypot( z->r, tmp ));
}

static void z_logbase(Rcomplex *z, Rcomplex *base)
{
    z_log(z);
    z_log(base);
    complex_div(z, base);
}

#define mod_iterate(n1, n2, i1, i2) for (i=i1=i2=0; i<n; \
    i1 = (++i1 == n1) ? 0 : i1,\
    i2 = (++i2 == n2) ? 0 : i2,\
    ++i)

static SEXP
externalVectorLog(SEXP x, SEXP base)
{
    SEXP x_storage = GET_SLOT(x, install("storage"));
    storageMethods_t* x_methods = getStorageMethods(x_storage);
    int nx = asInteger(GET_SLOT(x_storage, install("length")));
    SEXPTYPE mode = x_methods->mode;
    SEXP ans_storage;
    SEXP x_ptr, ans_ptr;
    int naflag = 0;
    int warn = 0;
    int x_warn = 0;
    int base_warn = 0;

    if (isExternalVectorWithStorage(base)) {
        SEXP base_storage = GET_SLOT(base, install("storage"));
        storageMethods_t* base_methods = getStorageMethods(base_storage);
        int nbase = asInteger(GET_SLOT(base_storage, install("length")));
        int n = nx >= nbase ? nx : nbase;
        storageMethods_t* ans_methods;
        setExtElt_t setElt;
        getExtElt_t base_getElt;
        getExtElt_t x_getElt;
        primitiveElt_t x_elt;
        primitiveElt_t base_elt;
        SEXP base_ptr;
        int i, ibase, ix;

        if (base_methods->mode == CPLXSXP)
            mode = CPLXSXP;
        if (mode == CPLXSXP) {
            x_getElt = x_methods->complexGetElt;
            base_getElt = base_methods->complexGetElt;
        } else {
            mode = REALSXP;
            x_getElt = x_methods->numericGetElt;
            base_getElt = base_methods->numericGetElt;
        }
        ans_storage = getArithmeticAnswerStorage(x_storage,
                                                 x_methods, nx,
                                                 base_storage,
                                                 base_methods,
                                                 nbase, mode);
        PROTECT(ans_storage);
        ans_methods = getStorageMethods(ans_storage);
        setElt = ans_methods->setExtElt;
        x_ptr = GET_STORAGE_EXT_PTR(x_storage);
        ans_ptr = GET_STORAGE_EXT_PTR(ans_storage);
        base_ptr = GET_STORAGE_EXT_PTR(base_storage);

        if (mode == REALSXP) {
            mod_iterate(nx, nbase, ix, ibase) {
                x_getElt(x_ptr, x_methods, ix, &x_elt, &x_warn);
                base_getElt(base_ptr, base_methods, ibase,
                            &base_elt, &base_warn);
                if (ISNAN(x_elt.numeric)) {
                    if (ISNA(base_elt.numeric))
                        x_elt.numeric = NA_REAL;
                } else {
                    if (ISNAN(base_elt.numeric))
                        x_elt.numeric = base_elt.numeric;
                    else {
                        x_elt.numeric = logbase(x_elt.numeric,
                                                base_elt.numeric);
                        if (ISNAN(x_elt.numeric))
                            naflag = 1;
                    }
                }
                setElt(ans_ptr, ans_methods, i, &x_elt, &warn);
            }
        } else {
            mod_iterate(nx, nbase, ix, ibase) {
                x_getElt(x_ptr, x_methods, ix, &x_elt, &x_warn);
                base_getElt(base_ptr, base_methods, ibase,
                            &base_elt, &base_warn);
                if (ISNA(x_elt.cmplx.r) && ISNA(x_elt.cmplx.i) &&
                    ISNA(base_elt.cmplx.r) && ISNA(base_elt.cmplx.i)) {
                } else {
                    z_logbase(&x_elt.cmplx, &base_elt.cmplx);
#ifndef IEEE_754
                    if (ISNA(base_elt.cmplx.r)) {
                        if (!ISNA(base_elt.cmplx.i))
                            base_elt.cmplx.i = NA_REAL;
                        naflag = 1;
                    } else if (ISNA(base_elt.cmplx.i)) {
                        base_elt.cmplx.r = NA_REAL;
                        naflag = 1;
                    }
#endif
                }
                setElt(ans_ptr, ans_methods, i, &base_elt, &warn);
            }
        }
    } else {
        SEXPTYPE base_mode = TYPEOF(base);
        int nbase = LENGTH(base);
        int n = nx >= nbase ? nx : nbase;
        storageMethods_t* ans_methods;
        setExtElt_t setElt;
        getExtElt_t x_getElt;
        primitiveElt_t x_elt;
        int i, ibase, ix;

        if (base_mode == CPLXSXP)
            mode = CPLXSXP;
        if (mode == CPLXSXP) {
            x_getElt = x_methods->complexGetElt;
        } else {
            mode = REALSXP;
            x_getElt = x_methods->numericGetElt;
        }
        ans_storage = x_methods->alloc(x_storage,
                                       allocVector(mode, 1), nx);
        PROTECT(ans_storage);
        ans_methods = getStorageMethods(ans_storage);
        setElt = ans_methods->setExtElt;
        x_ptr = GET_STORAGE_EXT_PTR(x_storage);
        ans_ptr = GET_STORAGE_EXT_PTR(ans_storage);

        if (mode == REALSXP) {
            if (base_mode == REALSXP) {
                double* base_ptr = REAL(base);
                mod_iterate(nx, nbase, ix, ibase) {
                    double base_elt = base_ptr[ibase];
                    x_getElt(x_ptr, x_methods, ix, &x_elt, &x_warn);
                    if (ISNAN(x_elt.numeric)) {
                        if (ISNA(base_elt))
                            x_elt.numeric = NA_REAL;
                    } else {
                        if (ISNAN(base_elt))
                            x_elt.numeric = base_elt;
                        else {
                            x_elt.numeric = logbase(x_elt.numeric,
                                                    base_elt);
                            if (ISNAN(x_elt.numeric))
                                naflag = 1;
                        }
                    }
                    setElt(ans_ptr, ans_methods, i, &x_elt, &warn);
                }
            } else { /* logical or integer */
                int* base_ptr = INTEGER(base);
                mod_iterate(nx, nbase, ix, ibase) {
                    int base_elt = base_ptr[ibase];
                    x_getElt(x_ptr, x_methods, ix, &x_elt, &x_warn);
                    if (base_elt == NA_INTEGER)
                        x_elt.numeric = NA_REAL;
                    else if (!ISNAN(x_elt.numeric)) {
                        x_elt.numeric = logbase(x_elt.numeric,
                                                (double) base_elt);
                        if (ISNAN(x_elt.numeric))
                            naflag = 1;
                    }
                    setElt(ans_ptr, ans_methods, i, &x_elt, &warn);
                }
            }
        } else {
            if (base_mode == CPLXSXP) {
                Rcomplex* base_ptr = COMPLEX(base);
                primitiveElt_t base_elt;
                mod_iterate(nx, nbase, ix, ibase) {
                    base_elt.cmplx.r = base_ptr[ibase].r;
                    base_elt.cmplx.i = base_ptr[ibase].i;
                    x_getElt(x_ptr, x_methods, ix, &x_elt, &x_warn);
                    if (ISNA(x_elt.cmplx.r) && ISNA(x_elt.cmplx.i) &&
                        ISNA(base_elt.cmplx.r) && ISNA(base_elt.cmplx.i)) {
                    } else {
                        z_logbase(&x_elt.cmplx, &base_elt.cmplx);
#ifndef IEEE_754
                        if (ISNA(base_elt.cmplx.r)) {
                            if (!ISNA(base_elt.cmplx.i))
                                base_elt.cmplx.i = NA_REAL;
                            naflag = 1;
                        } else if (ISNA(base_elt.cmplx.i)) {
                            base_elt.cmplx.r = NA_REAL;
                            naflag = 1;
                        }
#endif
                    }
                    setElt(ans_ptr, ans_methods, i, &base_elt, &warn);
                }
            } else if (base_mode == REALSXP) {
                double* base_ptr = REAL(base);
                mod_iterate(nx, nbase, ix, ibase) {
                    double base_elt = base_ptr[ibase];
                    x_getElt(x_ptr, x_methods, ix, &x_elt, &x_warn);
                    if (ISNA(x_elt.cmplx.r) && ISNA(x_elt.cmplx.i) &&
                        ISNA(base_elt)) {
                    } else {
                        base_elt = R_log(base_elt);
#ifdef IEEE_754
                        z_log(&x_elt.cmplx);
                        x_elt.cmplx.r /= base_elt;
                        x_elt.cmplx.r /= base_elt;
#else
                        if (base_elt == 0.0) {
                            base_elt.cmplx.r = NA_REAL;
                            x_elt.cmplx.i = NA_REAL;
                            naflag = 1;
                        } else {
                            z_log(&x_elt.cmplx);
                            x_elt.cmplx.r /= base_elt;
                            x_elt.cmplx.r /= base_elt;
                            if (ISNA(x_elt.cmplx.r)) {
                                if (!ISNA(x_elt.cmplx.i))
                                    x_elt.cmplx.i = NA_REAL;
                                naflag = 1;
                            } else if (ISNA(x_elt.cmplx.i)) {
                                base_elt.cmplx.r = NA_REAL;
                                naflag = 1;
                            }
                        }
#endif
                    }
                    setElt(ans_ptr, ans_methods, i, &x_elt, &warn);
                }
            } else {
                int* base_ptr = INTEGER(base);
                mod_iterate(nx, nbase, ix, ibase) {
                    int base_elt = base_ptr[ibase];
                    x_getElt(x_ptr, x_methods, ix, &x_elt, &x_warn);
                    if (ISNA(x_elt.cmplx.r) && ISNA(x_elt.cmplx.i) &&
                        base_elt == NA_INTEGER) {
                    } else {
                        double base_log = R_log((double) base_elt);
#ifdef IEEE_754
                        z_log(&x_elt.cmplx);
                        x_elt.cmplx.r /= (double) base_log;
                        x_elt.cmplx.r /= (double) base_log;
#else
                        if (base_log == 0) {
                            x_elt.cmplx.r = NA_REAL;
                            x_elt.cmplx.i = NA_REAL;
                            naflag = 1;
                        } else {
                            z_log(&x_elt.cmplx);
                            x_elt.cmplx.r /= base_log;
                            x_elt.cmplx.r /= base_log;
                            if (ISNA(x_elt.cmplx.r)) {
                                if (!ISNA(x_elt.cmplx.i))
                                    x_elt.cmplx.i = NA_REAL;
                                naflag = 1;
                            } else if (ISNA(x_elt.cmplx.i)) {
                                x_elt.cmplx.r = NA_REAL;
                                naflag = 1;
                            }
                        }
#endif
                    }
                    setElt(ans_ptr, ans_methods, i, &x_elt, &warn);
                }
            }
        }
    }
    CoercionWarning(x_warn);
    CoercionWarning(base_warn);
    CoercionWarning(warn);
    if (naflag)
        warning("NAs produced in function \"log\"");
    UNPROTECT(1);
    return ans_storage;
}

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(externalVectorLog, 2),
    {NULL, NULL, 0},
};

const R_CallMethodDef*
math_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}
