#include <Rmath.h>
#include "subscript.h"
#include "coerce.h"

SEXP positiveIndex;
SEXP noneIndex;
SEXP allIndex;
SEXP sliceIndex;
SEXP matrixIndex;
SEXP NAIndex;

const SEXPREC* getIndexType(const SEXPREC* indx)
{
    const char* klass;
    if (indx == noneIndex || indx == allIndex || !isObject((SEXP)indx))
        return indx;
    klass = CHAR(STRING_ELT(GET_CLASS((SEXP)indx), 0));
    switch (klass[0]) {
    case 'n':
        if (strcmp(klass+1, "oneIndex") == 0)
            return noneIndex;
        break;
    case 'a':
        if (strcmp(klass+1, "llIndex") == 0)
            return allIndex;
        break;
    case 's':
        if (strcmp(klass+1, "liceIndex") == 0)
            return sliceIndex;
        break;
    case 'N':
        if (strcmp(klass+1, "AIndex") == 0)
            return NAIndex;
        break;
    case 'p':
        if (strcmp(klass+1, "ositiveIndex") == 0)
            return positiveIndex;
        break;
    default:
        break;
    }
    return indx;
}

static SEXP
initPositiveIndex(SEXP args)
{
    SEXP x;
    int n;
    int* xrange;
    int xmin, xmax;
    int noNA = 1;

    for (args = CDR(args); args != R_NilValue; args = CDR(args)) {
        if (TAG(args) == R_NilValue)
            break;
    }
    if (args == R_NilValue) {
        x = allocVector(INTSXP, 0);
        n = 0;
    } else {
        x = CAR(args);
        if (TYPEOF(x) == INTSXP) {
            SEXP tmp;
            SEXP nm = getAttrib(x, R_NamesSymbol);
            n = LENGTH(x);
            tmp = allocVector(INTSXP, n);
            memcpy(INTEGER(tmp), INTEGER(x), n*sizeof(int));
            x = tmp;
            setAttrib(x, R_NamesSymbol, nm);
        } else {
            x = coerceVector(x, INTSXP);
            n = LENGTH(x);
        }
    }
    PROTECT(x);
    if (n == 0) {
        xmin = xmax = 0;
    } else {
        int* ix = INTEGER(x);
        int i;
        xmax = NA_INTEGER;
        xmin = INT_MAX;
        for (i = 0; i < n; i++) {
            int tmp = ix[i];
            if (tmp == NA_INTEGER) {
                noNA = 0;
            } else {
                if (tmp > xmax)
                    xmax = tmp;
                if (tmp < xmin)
                    xmin = tmp;
            }
        }
        if (xmax == NA_INTEGER)
            xmin = NA_INTEGER;
    }
    SET_CLASS(x, GET_CLASS(positiveIndex));
    SET_SLOT(x, install("rangeIndex"), allocVector(INTSXP, 2));
    xrange = INTEGER(GET_SLOT(x, install("rangeIndex")));
    xrange[0] = xmin;
    xrange[1] = xmax;
    SET_SLOT(x, install("noNA"), ScalarLogical(noNA));
    UNPROTECT(1);
    return x;
}

static SEXP initAllIndex() { return allIndex; }
static SEXP initNoneIndex() { return noneIndex; }
static SEXP initNAIndex(SEXP len)
{
    SEXP ans = duplicate(NAIndex);
    int n = asInteger(len);
    INTEGER(GET_SLOT(ans, install("Length")))[0] = n;
    return ans;
}

static SEXP
makeSlice(SEXP sStart, SEXP sLength, SEXP sStride)
{
    SEXP slice;
    double n = asReal(sLength);
    double start;
    double stride = 0;
    double end;
    double* content;

    if (ISNAN(n) || n < 0.0)
        error("invalid length value");
    if (n < 1.0)
        return noneIndex;
    /* n is greater than one: floor is same as truncation */
    n = floor(n);

    start = asReal(sStart);
    if (!R_FINITE(start))
        error("invalid argument for start");
    start = ftrunc(start);
    if (start < 0)
        error("can not create slice with negative numbers");
    if (n > 1) {
        stride = asReal(sStride);
        if (!R_FINITE(stride))
            error("invalid argument for stride");
        stride = ftrunc(stride);
        if (stride == 0 && start == 0)
            return noneIndex;
        end = start + (n-1)*stride;
        if (n > 1 && end < 0)
            error("start and end have opposite signs");
    }
    if (n == 0) {
        return noneIndex;
    } else if (n == 1) {
        stride = 0;
    }

    slice = duplicate(sliceIndex);
    content = REAL(GET_SLOT(slice, install("content")));
    content[0] = start;
    content[1] = n;
    content[2] = stride;
    return slice;
}

static SEXP
makeSliceFromLogicalIndex(SEXP sIndx)
{
    int nindex = LENGTH(sIndx);
    int* indx;
    Rboolean first;

    if (nindex == 0)
        return noneIndex;
    indx = LOGICAL(sIndx);
    first = indx[0];
    if (nindex > 1) {
        Rboolean current = 0;
        int i;
        for (i = 1; i < nindex; i++) {
            current = indx[i];
            if (current != first)
                break;
        }
        if (current != first) {
            SEXP slice;
            double* content;
            if (first == NA_LOGICAL || current == NA_LOGICAL)
                return sIndx;
            if (nindex == i+1) {
                slice = duplicate(sliceIndex);
                content = REAL(GET_SLOT(slice, install("content")));

                if (nindex == 2) {
                    content[0] = (first == FALSE)?2.0:1.0;
                    content[1] = R_PosInf;
                    content[2] = 2.0;
                    return slice;
                }
                content[0] = (double) nindex;
                content[1] = R_PosInf;
                content[2] = nindex;
                if (first != FALSE)
                    LOGICAL(GET_SLOT(slice, install("negative")))[0] = TRUE;
            } else if (i == 1) {
                int stride = 0;
                for (i++; i < nindex; i++) {
                    current = indx[i];
                    if (current == NA_LOGICAL)
                        return sIndx;
                    if (stride == 0) {
                        if (current == first)
                            stride = i;
                    } else if (i % stride == 0 && current != first)
                        return sIndx;
                }

                slice = duplicate(sliceIndex);
                content = REAL(GET_SLOT(slice, install("content")));
                content[1] = R_PosInf;
                if (stride == 2) {
                    content[0] = (first == FALSE)?2.0:1.0;
                    content[2] = 2.0;
                } else {
                    content[0] = 1.0;
                    content[2] = (double) stride;
                    if (first == FALSE)
                        LOGICAL(GET_SLOT(slice, install("negative")))[0] = TRUE;
                }
            } else {
                int start = i;
                int stride = 0;
                for (i++; i < nindex; i++) {
                    current = indx[i];
                    if (current == NA_LOGICAL)
                        return sIndx;
                    if (stride == 0) {
                        if (current != first) {
                            if (i <= 2*start)
                                return sIndx;
                            stride = i-start;
                        }
                    } else if ((i-start) % stride == 0 && current != first)
                        return sIndx;
                }
                
                slice = duplicate(sliceIndex);
                content = REAL(GET_SLOT(slice, install("content")));
                content[0] = (double) (start+1);
                content[1] = R_PosInf;
                content[2] = (double) stride;
                if (first != FALSE)
                    LOGICAL(GET_SLOT(slice, install("negative")))[0] = TRUE;
            }
            return slice;
        }
    }

    if (first == FALSE)
        return noneIndex;
    if (first == TRUE)
        return allIndex;
    else {
        SEXP slice = duplicate(sliceIndex);
        double* content = REAL(GET_SLOT(slice, install("content")));

        content[0] = NA_REAL;
        content[1] = 1.0;
        content[2] = 0.0;
        return slice;
    }
}

SEXP
slimplifyPositiveSubscript(const SEXPREC* s, subs_NAaction_t NAaction)
{
    int i;
    int ns = LENGTH((SEXP)s);
    int* sub = INTEGER((SEXP)s);
    int start, end, stride;
    double* content;
    SEXP ans = R_NilValue;
    int warn = 0;

    if (ns == 0)
        return noneIndex;
    if (ns == 1) {
        ans = duplicate(sliceIndex);
        content = REAL(GET_SLOT(ans, install("content")));
        content[0] = RealFromInteger(sub[0], &warn);
        content[1] = 1.0;
        content[2] = 0.0;
        content[3] = 1.0;
        return ans;
    }
    start = sub[0];
    end = sub[ns-1];
    switch (NAaction) {
    case REMOVE_NA:
        error("Na's should be removed before calling silmlifyPositiveSubscript");
    default:
    case HAVE_NA:
        ans = duplicate((SEXP)s);
        PROTECT(ans);
        stride = 0;
        break;
    case NO_NA:
        stride = (double) (end - start)/(ns-1);
        for (i = ns-2; i > 0; i--) {
            int indx = sub[i];
            if (stride != end-indx) {
                ans = duplicate((SEXP)s);
                PROTECT(ans);
                break;
            }
            end = indx;
        }
        break;
    case KEEP_NA:
        if (start == NA_INTEGER || end == NA_INTEGER) {
            ans = duplicate((SEXP)s);
            PROTECT(ans);
            stride = 0;
        } else {
            stride = (end-start)/(ns-1);
            for (i = ns-2; i > 0; i++) {
                int indx = sub[i];
                if (indx == NA_INTEGER || stride != end-indx) {
                    ans = duplicate((SEXP)s);
                    PROTECT(ans);
                    break;
                }
                end = indx;
            }
        }
        break;
    }
    if (ans != R_NilValue) {
        SET_CLASS(ans, mkString("positiveIndex"));
    } else {
        ans = duplicate(sliceIndex);
        content = REAL(GET_SLOT(ans, install("content")));
        content[0] = (double) start;
        content[1] = (double) ns;
        content[2] = (double) stride;
        content[3] = (double) stride;
    }
    UNPROTECT(1);
    return ans;
}

#define SMALL_SUBSCRIPT_SIZE 5

Rboolean
isSimpleSubscript(const SEXPREC* s, int nx, subs_NAaction_t NAaction,
                  int stretch)
{
    SEXPTYPE mode = TYPEOF((SEXP) s);
    int ns = LENGTH((SEXP) s);

    if (ns == 0)
        return TRUE;

    if ((mode == REALSXP || mode == INTSXP) && ns <= SMALL_SUBSCRIPT_SIZE) {
        if (mode == REALSXP) {
            double* sub = REAL((SEXP) s);
            int i;
            for (i = 0; i < ns; i++) {
                double tmp = sub[i];
                if (!R_finite(tmp) || tmp < 1 || tmp >= nx+1)
                    return FALSE;
            }
        } else {
            int* sub = INTEGER((SEXP) s);
            int i;
            for (i = 0; i < ns; i++) {
                int tmp = sub[i];
                if (tmp == NA_INTEGER || tmp < 1 || tmp >= nx+1)
                    return FALSE;
            }
        }
        return TRUE;
    }
    return FALSE;
}

SEXP
simplifyProcessedSubscript(SEXP s, subs_NAaction_t NAaction,
                           int stretch, int startIndex, int ntotal)
{
    SEXPTYPE mode = TYPEOF(s);
    int i;
    int ns = LENGTH(s);
    int start, end, stride;
    double* content;
    SEXP ans = R_NilValue;

    if (ns == 0)
        return noneIndex;

    if (ns == 1) {
        ans = duplicate(sliceIndex);
        content = REAL(GET_SLOT(ans, install("content")));
        content[0] = asReal(s);
        content[0] -= startIndex;
        content[1] = 1.0;
        content[2] = 0.0;
        return ans;
    }
    start = asInteger(s);
    end = (mode == REALSXP)?((int)REAL(s)[ns-1]):INTEGER(s)[ns-1];
    switch (NAaction) {
    case REMOVE_NA:
        error("Na's should be removed before calling silmlifyPositiveSubscript");
    default:
    case HAVE_NA:
        ans = s;
        stride = 0;
        break;
    case NO_NA:
        stride = (double) (end - start)/(ns-1);
        if (stride*(ns-1) != end-start) {
            ans = s;
        } else if (mode == REALSXP) {
            double* sub = REAL(s);
            for (i = ns-2; i > 0; i--) {
                int indx = (int) sub[i];
                if (stride != end-indx) {
                    ans = s;
                    break;
                }
                end = indx;
            }
        } else {
            int* sub = INTEGER(s);
            for (i = ns-2; i > 0; i--) {
                int indx = sub[i];
                if (stride != end-indx) {
                    ans = s;
                    break;
                }
                end = indx;
            }
        }
        break;
    case KEEP_NA:
        if (start == NA_INTEGER || end == NA_INTEGER) {
            ans = s;
            stride = 0;
        } else if (mode == REALSXP) {
            double* sub = REAL(s);
            stride = (end-start)/(ns-1);
            if (stride*(ns-1) != end-start)
                ans = s;
            else {
                for (i = ns-2; i > 0; i++) {
                    double tmp = sub[i];
                    int indx = (int) tmp;
                    if (!R_finite(tmp) || stride != end-indx) {
                        ans = s;
                        break;
                    }
                    end = indx;
                }
            }
        } else {
            int* sub = INTEGER(s);
            stride = (end-start)/(ns-1);
            if (stride*(ns-1) != end-start)
                ans = s;
            else {
                for (i = ns-2; i > 0; i++) {
                    int indx = sub[i];
                    if (indx == NA_INTEGER || stride != end-indx) {
                        ans = s;
                        break;
                    }
                    end = indx;
                }
            }
        }
        break;
    }
    if (ans == s) {
        int noNA = 1;
        int smin = INT_MAX;
        int smax = NA_INTEGER;
        int* sub;
        ans = mode==INTSXP?duplicate(s):coerceVector(s, INTSXP);
        sub = INTEGER(ans);
        PROTECT(ans);
        switch (NAaction) {
        case KEEP_NA:
        case HAVE_NA:
            if (startIndex == 0) {
                for (i = 0; i < ns; i++)
                    if (sub[i] == NA_INTEGER) {
                        noNA = 0;
                    } else {
                        int tmp = sub[i];
                        if (tmp > smax) {
                            smax = tmp;
                        }
                        if (tmp < smin)
                            smin = tmp;
                    }
            } else if (startIndex == 1) {
                for (i = 0; i < ns; i++)
                    if (sub[i] == NA_INTEGER) {
                        noNA = 0;
                    } else {
                        int tmp = --sub[i];
                        if (tmp > smax) {
                            smax = tmp;
                        }
                        if (tmp < smin)
                            smin = tmp;
                    }
            } else {
                for (i = 0; i < ns; i++)
                    if (sub[i] == NA_INTEGER) {
                        noNA = 0;
                    } else {
                        int tmp = sub[i] -= startIndex;
                        if (tmp > smax) {
                            smax = tmp;
                        }
                        if (tmp < smin)
                            smin = tmp;
                    }
            }
            break;
        case NO_NA:
            if (startIndex == 0) {
                for (i = 0; i < ns; i++) {
                    int tmp = sub[i];
                    if (tmp > smax) {
                        smax = tmp;
                    }
                    if (tmp < smin)
                        smin = tmp;
                }
            } else if (startIndex == 1) {
                for (i = 0; i < ns; i++) {
                    int tmp = --sub[i];
                    if (tmp > smax) {
                        smax = tmp;
                    }
                    if (tmp < smin)
                        smin = tmp;
                }
            } else {
                for (i = 0; i < ns; i++) {
                    int tmp = sub[i] -= startIndex;
                    if (tmp > smax) {
                        smax = tmp;
                    }
                    if (tmp < smin)
                        smin = tmp;
                }
            }
            break;
        default:
        case REMOVE_NA:
            /* can not reach*/
            break;
        }
        SET_CLASS(ans, mkString("positiveIndex"));
        SET_SLOT(ans, install("rangeIndex"), allocVector(INTSXP, 2));
        sub = INTEGER(GET_SLOT(ans, install("rangeIndex")));
        sub[1] = smax;
        sub[0] = smax==NA_INTEGER?NA_INTEGER:smin;
        SET_SLOT(ans, install("noNA"), ScalarLogical(noNA));
    } else if (ns == ntotal && start == startIndex && stride == 1) {
        return allIndex;
    } else {
        ans = duplicate(sliceIndex);
        PROTECT(ans);
        content = REAL(GET_SLOT(ans, install("content")));
        content[0] = (double) (start-startIndex);
        content[1] = (double) ns;
        content[2] = (double) stride;
    }
    UNPROTECT(1);
    return ans;
}

static SEXP
subsetPrimitiveBySlice(SEXP x, SEXP slice)
{
    double* content;
    SEXPTYPE mode = TYPEOF(x);
    SEXP ans;
    int n, start, stride, i;
    int len, end;

    if (!isVector(x))
        error("can not subset non-vector");
    len = LENGTH(x);
    content = REAL(GET_SLOT(slice, install("content")));

    start = (int) content[0];
    n = (int) content[1];
    stride = (int) content[2];
    end = start + (n-1)*stride;
    if (start < 0 || start >= len)
        error("invalid slice");
    if (end < 0 || end >= len)
        error("invalid slice");

    ans = allocVector(mode, n);
    switch (mode) {
    case LGLSXP:
    {
        int* src = LOGICAL(x);
        int* dest = LOGICAL(ans);
        for (i = 0; i < n; i++, start += stride)
            dest[i] = src[start];
        break;
    }
    case INTSXP:
    {
        int* src = INTEGER(x);
        int* dest = INTEGER(ans);
        for (i = 0; i < n; i++, start += stride)
            dest[i] = src[start];
        break;
    }
    case REALSXP:
    {
        double* src = REAL(x);
        double* dest = REAL(ans);
        for (i = 0; i < n; i++, start += stride)
            dest[i] = src[start];
        break;
    }
    case CPLXSXP:
    {
        Rcomplex* src = COMPLEX(x);
        Rcomplex* dest = COMPLEX(ans);
        for (i = 0; i < n; i++, start += stride) {
            dest[i].r = src[start].r;
            dest[i].i = src[start].i;
        }
        break;
    }
    case STRSXP:
    {
        for (i = 0; i < n; i++, start += stride)
            SET_STRING_ELT(ans, i, STRING_ELT(x, start));
        break;
    }
    case VECSXP:
    case EXPRSXP:
    {
        for (i = 0; i < n; i++, start += stride)
            SET_VECTOR_ELT(ans, i, VECTOR_ELT(x, start));
        break;
    }
    }
    return ans;
}

static SEXP
subsetPrimitiveByPositiveIndex(SEXP x, SEXP indx)
{
    SEXPTYPE mode = TYPEOF(x);
    SEXP ans;
    int n = LENGTH(indx);
    int noNA = asLogical(GET_SLOT(indx, install("noNA")));
    int* rangeIndx = INTEGER(GET_SLOT(indx, install("maxIndex")));
    int minIndx = rangeIndx[0];
    int maxIndx = rangeIndx[1];
    int checkIndx;
    int* indxptr = INTEGER(indx);
    int len, i;

    if (!isVector(x))
        error("can not subset non-vector");
    len = LENGTH(x);
    checkIndx = (maxIndx == NA_INTEGER) || (minIndx == NA_INTEGER) ||
        (maxIndx >= len) || (minIndx < 0) || !noNA;

    ans = allocVector(mode, n);
    switch (mode) {
#define SUBSET_VEC(NA)                                      \
        if (checkIndx) {                                   \
            for (i = 0; i < n; i++) {                       \
                int j = indxptr[i];                         \
                if (j == NA_INTEGER || j < 0 || j >= len) { \
                    SET_NA(dest, i, NA);                    \
                } else {                                    \
                    SET_SCALAR(dest, i, src, j);            \
                }                                           \
            }                                               \
        } else {                                            \
            for (i = 0; i < n; i++) {                       \
                int j = indxptr[i];                         \
                SET_SCALAR(dest, i, src, j);                \
            }                                               \
        }
#define SET_SCALAR(d, i, s, j) (d)[i] = (s)[j]
#define SET_NA(d, i, na) (d)[i] = (na)
    case LGLSXP:
    {
        int* src = LOGICAL(x);
        int* dest = LOGICAL(ans);
        SUBSET_VEC(NA_LOGICAL);
        break;
    }
    case INTSXP:
    {
        int* src = INTEGER(x);
        int* dest = INTEGER(ans);
        SUBSET_VEC(NA_INTEGER);
        break;
    }
    case REALSXP:
    {
        double* src = REAL(x);
        double* dest = REAL(ans);
        SUBSET_VEC(NA_REAL);
        break;
    }
#undef SET_SCALAR
#define SET_SCALAR(d, i1, s, j) (d)[i1].r = (s)[j].r; (d)[i1].i = (s)[j].i
#undef SET_NA
#define SET_NA(d, i1, na) (d)[i1].r = (na); (d)[i1].i = (na)
    case CPLXSXP:
    {
        Rcomplex* src = COMPLEX(x);
        Rcomplex* dest = COMPLEX(ans);
        SUBSET_VEC(NA_REAL);
        break;
    }
#undef SUBSET_VEC
#define SUBSET_VEC(NA)                                      \
        if (checkIndx) {                                   \
            for (i = 0; i < n; i++) {                       \
                int j = indxptr[i];                         \
                if (j == NA_INTEGER || j < 0 || j >= len) { \
                    SET_NA(ans, i, NA);                     \
                } else {                                    \
                    SET_SCALAR(ans, i, x, j);               \
                }                                           \
            }                                               \
        } else {                                            \
            for (i = 0; i < n; i++) {                       \
                int j = indxptr[i];                         \
                SET_SCALAR(ans, i, x, j);                   \
            }                                               \
        }
#undef SET_SCALAR
#define SET_SCALAR(d, i, s, j) SET_STRING_ELT((d), (i), STRING_ELT((s), (j)))
#undef SET_NA
#define SET_NA(d, i, na) SET_STRING_ELT((d), (i), (na))
    case STRSXP:
    {
        SUBSET_VEC(NA_STRING);
        break;
    }
#undef SET_SCALAR
#define SET_SCALAR(d, i, s, j) SET_VECTOR_ELT((d), (i), VECTOR_ELT((s), (j)))
#undef SET_NA
#define SET_NA(d, i, na) SET_VECTOR_ELT((d), (i), (na))
    case VECSXP:
    case EXPRSXP:
    {
        SUBSET_VEC(R_NilValue);
        break;
    }
#undef SUBSET_VEC
#undef SET_SCALAR
#undef SET_NA
    default:
        break;
    }
    return ans;
}

int
getIndexLength(SEXP indx, int n)
{
    const SEXPREC* type = getIndexType(indx);
    int warn = 0;
    if (type == allIndex) {
        return n;
    } else if (type == noneIndex) {
        return 0;
    } else if (type == NAIndex) {
        return asInteger(GET_SLOT(indx, install("Length")));
    } else if (type == sliceIndex) {
        n = IntegerFromReal(REAL(GET_SLOT(indx, install("content")))[1],
                            &warn);
        CoercionWarning(warn);
        return n;
    } else if (type == positiveIndex) {
        return LENGTH(indx);
    } else if (type == matrixIndex) {
        int* dim = INTEGER(GET_SLOT(indx, install("origdim")));
        return (getIndexLength(GET_SLOT(indx, install("row")), dim[0])*
                getIndexLength(GET_SLOT(indx, install("col")), dim[1]));
    } else {
        error("unknown index type");
        return 0; /* -Wall */
    }
}

SEXP
asPositiveIndex(SEXP indx, int n)
{
    const SEXPREC* type = getIndexType(indx);
    if (n != getIndexLength(indx, n))
        error("index length mismatch");
    if (type == allIndex) {
        SEXP ans;
        int i;
        ans = allocVector(INTSXP, n);
        for (i = 0; i < n; i++)
            INTEGER(ans)[i] = i;
        return ans;
    } else if (type == noneIndex) {
        return allocVector(INTSXP, 0);
    } else if (type == NAIndex) {
        SEXP ans;
        int i;
        ans = allocVector(INTSXP, n);
        for (i = 0; i < n; i++)
            INTEGER(ans)[i] = NA_INTEGER;
        return ans;
    } else if (type == sliceIndex) {
        double* content = REAL(GET_SLOT(indx, install("content")));
        SEXP ans;
        int i;
        int start = content[0];
        int stride = content[2];
        ans = allocVector(INTSXP, n);
        for (i = 0; i < n; i++, start += stride)
            INTEGER(ans)[i] = start;
        return ans;
    } else if (type == positiveIndex) {
        return indx;
    } else {
        error("unknown index type");
        return R_NilValue; /* -Wall */
    }
}

static SEXP
matrixToVectorIndex(SEXP rowindx, SEXP rowlen,
                    SEXP colindx, SEXP collen, SEXP origdims,
                    SEXP sub, SEXP sublen)
{
    int n = INTEGER(origdims)[0];
    int p = INTEGER(origdims)[1];
    int ntotal = n*INTEGER(origdims)[1];
    int nrow = asInteger(rowlen);
    int ncol = asInteger(collen);
    int nsub = asInteger(sublen);
    subs_NAaction_t NAaction;
    SEXP ans;
    const SEXPREC* rowtype = getIndexType(rowindx);
    const SEXPREC* coltype = getIndexType(colindx);
    const SEXPREC* subtype = getIndexType(sub);
    if (nrow == 0 || ncol == 0 || nsub == 0)
        return noneIndex;
    if (rowtype == NAIndex || coltype == NAIndex || subtype == NAIndex) {
        ans = duplicate(NAIndex);
        INTEGER(GET_SLOT(ans, install("length")))[0] =
            (nsub == NA_INTEGER)?nrow*ncol:nsub;
        return ans;
    }
    if (subtype == allIndex &&
        rowtype == sliceIndex && coltype == sliceIndex
        && nrow == n && ncol == p) {
        double* content = REAL(GET_SLOT(rowindx, install("content")));
        int start = (int) content[0];
        int stride = (int) content[2];
        if (start == 0 && stride == 1) {
            content = REAL(GET_SLOT(colindx, install("content")));
            start = (int) content[0];
            stride = (int) content[2];
            if (start == 0 && stride == 1) {
                ans = duplicate(sliceIndex);
                content = REAL(GET_SLOT(ans, install("content")));
                content[0] = 0.0;
                content[1] = (double) (nrow*ncol);
                content[2] = 1.0;
                return ans;
            }
        }
    }
    if (subtype == allIndex) {
        int i;
        int* ians;
        nsub = nrow*ncol;
        ans = allocVector(INTSXP, nsub);
        PROTECT(ans);
        ians = INTEGER(ans);
        if (rowtype == sliceIndex) {
            double* content = REAL(GET_SLOT(rowindx, install("content")));
            int start = (int) content[0];
            int stride = (int) content[2];
            for (i = 0; i < nrow; i++, start += stride)
                ians[i] = start;
        } else memcpy(ians, INTEGER(rowindx), nrow*sizeof(int));
        if (coltype == sliceIndex) {
            double* content = REAL(GET_SLOT(colindx, install("content")));
            int start = (int) content[0];
            int stride = (int) content[2];
            start *= n;
            stride *= n;
            for (i = 1, start += stride; i < ncol; i++, start += stride) {
                int j, jj;
                for (j = 0, jj=i*nrow; j < nrow; j++, jj++)
                    ians[jj] = ians[j]+start;
            }
            start = (int) content[0];
            start *= n;
            for (i = 0; i < nrow; i++)
                ians[i] += start;
        } else {
            int* icol = INTEGER(colindx);
            for (i = ncol-1; i >= 0; i--) {
                int j, jj;
                int start = icol[i]*n;
                for (j = 0, jj=i*nrow; j < nrow; j++, jj++)
                    ians[jj] = ians[j]+start;
            }
        }
        NAaction = NO_NA;
    } else {
        int nmat = nrow*ncol;
        int* ians;
        int* isub;
        int i;
        ans = allocVector(INTSXP, nsub);
        PROTECT(ans);
        ians = INTEGER(ans);
        if (subtype == sliceIndex) {
            double* content = REAL(GET_SLOT(sub, install("content")));
            int start = (int) content[0];
            int stride = (int) content[2];
            isub = Calloc(nsub, int);
            for (i = 0; i < nsub; i++, start += stride)
                isub[i] = start;
            isub = ians;
        } else {
            isub = INTEGER(sub);
        }
        NAaction = NO_NA;
        if (rowtype == sliceIndex) {
            double* content = REAL(GET_SLOT(rowindx, install("content")));
            int start = (int) content[0];
            int stride = (int) content[2];
            for (i = 0; i < nsub; i++) {
                int j = isub[i];
                if (j < 0 || j > nmat || j == NA_INTEGER) {
                    ians[i] = NA_INTEGER;
                    NAaction = HAVE_NA;
                } else ians[i] = (j % nrow)*stride+start;
            }
        } else {
            int* irow = INTEGER(rowindx);
            for (i = 0; i < nsub; i++) {
                int j = isub[i];
                if (j < 0 || j > nmat || j == NA_INTEGER) {
                    ians[i] = NA_INTEGER;
                    NAaction = HAVE_NA;
                } else ians[i] = irow[isub[i] % nrow];
            }
        }
        if (coltype == sliceIndex) {
            double* content = REAL(GET_SLOT(colindx, install("content")));
            int start = (int) content[0];
            int stride = (int) content[2];
            start *= n;
            stride *= n;
            for (i = 0; i < nsub; i++) {
                if (ians[i] != NA_INTEGER)
                    ians[i] += (isub[i] / nrow)*stride+start;
            }
        } else {
            int* icol = INTEGER(rowindx);
            for (i = 0; i < nsub; i++) {
                if (ians[i] != NA_INTEGER)
                    ians[i] += n*icol[isub[i] / nrow];
            }
        }
        if (subtype == sliceIndex)
            Free(isub);
    }
    ans = simplifyProcessedSubscript(ans, NAaction, 0, 0, ntotal);
    UNPROTECT(1);
    return ans;
}

#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(subsetPrimitiveBySlice, 2),
    CALL_DEF(subsetPrimitiveByPositiveIndex, 2),
    CALL_DEF(makeSlice, 3),
    CALL_DEF(makeSliceFromLogicalIndex, 1),
    CALL_DEF(matrixToVectorIndex, 7),
    CALL_DEF(initAllIndex, 0),
    CALL_DEF(initNoneIndex, 0),
    CALL_DEF(initNAIndex, 1),
    {NULL, NULL, 0},
};

static const R_ExternalMethodDef R_ExternalDef  [] = {
    EXTERNAL_DEF(initPositiveIndex),
    {NULL, NULL, 0},
};

#ifdef MAKE_CLASS
#undef MAKE_CLASS
#define MAKE_CLASS(nm) R_do_MAKE_CLASS((char *) (nm))
#else
#define MAKE_CLASS(nm) MAKE_CLASS((char *) (nm))
#endif

const R_CallMethodDef*
slice_Call_methods(int* nptr)
{
    noneIndex = NEW_OBJECT(MAKE_CLASS("noneIndex"));
    R_PreserveObject(noneIndex);
    SET_NAMED(noneIndex, 2);

    allIndex = NEW_OBJECT(MAKE_CLASS("allIndex"));
    R_PreserveObject(allIndex);
    SET_NAMED(allIndex, 2);

    sliceIndex = NEW_OBJECT(MAKE_CLASS("sliceIndex"));
    R_PreserveObject(sliceIndex);
    SET_NAMED(sliceIndex, 2);

    NAIndex = NEW_OBJECT(MAKE_CLASS("NAIndex"));
    R_PreserveObject(NAIndex);
    SET_NAMED(NAIndex, 2);

    positiveIndex = NEW_OBJECT(MAKE_CLASS("positiveIndex"));
    R_PreserveObject(positiveIndex);
    SET_NAMED(positiveIndex, 2);

    matrixIndex = NEW_OBJECT(MAKE_CLASS("matrixIndex"));
    R_PreserveObject(matrixIndex);
    SET_NAMED(matrixIndex, 2);

    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}


const R_ExternalMethodDef*
slice_External_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_ExternalDef);
    return &R_ExternalDef[0];
}
